#include <conio.h>
#include "sin.h"
#include "cos.h"

#define PERSPECT   0//0x28F
#define CORRECT    54613
#define NUMPOINTS  5

extern "C" TRIANGLE(unsigned x1, unsigned y1,
                    unsigned x2, unsigned y2,
                    unsigned x3, unsigned y3,
                    unsigned color, unsigned page);

extern "C" long SQRT(long arg);

unsigned page = 0;

class TPoint
{
  public:
    int x, y, z;
    TPoint(int x, int y, int z)
    {
      TPoint::x = x;
      TPoint::y = y;
      TPoint::z = z;
    }
};

class TNormal
{
  public:
    long x, y, z;
    TNormal(long x, long y, long z)
    {
      TNormal::x = x;
      TNormal::y = y;
      TNormal::z = z;
    }
    TNormal(TPoint *p1, TPoint *p2, TPoint *p3)
    {
      long register t;
      TNormal n1(p2 -> x - p1 -> x, p2 -> y - p1 -> y, p2 -> z - p1 -> z),
              n2(p3 -> x - p1 -> x, p3 -> y - p1 -> y, p3 -> z - p1 -> z);

      x = n1.y * n2.z - n1.z * n2.y;
      y = n1.z * n2.x - n1.x * n2.z;
      z = n1.x * n2.y - n1.y * n2.x;

      t = (long)SQRT(x * x + y * y + z * z);

      x <<= 16;
      y <<= 16;
      z <<= 16;

      x /= t;
      y /= t;
      z /= t;
    }
};

long IMul(op1, op2)
{
  return (op1 >> 16) * op2;
}

void Convert3D(TPoint p, int &xx, int &yy)
{
  long register t = (long)p.z * PERSPECT;
  long register t_x = p.x, t_y = p.y;

  if (p.z > 0)
  {
    t_x *= t + 0x10000L;
    t_y *= t + 0x10000L;
    xx = int(t_x >> 16);
    yy = int(t_y >> 16);
  }
  else
  {
    t = -t;
    t_x <<= 16;
    t_y <<= 16;
    t_x /= t + 0x10000L;
    t_y /= t + 0x10000L;
    xx = (int)t_x;
    yy = (int)t_y;
  }
}

void Triangle2D(TPoint p1, TPoint p2, TPoint p3,
                unsigned x, unsigned y, unsigned color)
{
  int t_x1, t_y1, t_x2, t_y2, t_x3, t_y3;

  Convert3D(p1, t_x1, t_y1);
  Convert3D(p2, t_x2, t_y2);
  Convert3D(p3, t_x3, t_y3);

  t_y1 = (t_y1 * CORRECT) >> 16;
  t_y2 = (t_y2 * CORRECT) >> 16;
  t_y3 = (t_y3 * CORRECT) >> 16;

  TRIANGLE(t_x1 + x, t_y1 + y,
           t_x2 + x, t_y2 + y,
           t_x3 + x, t_y3 + y,
           color, page);
}

class TTriangle {
    TPoint *m1, *m2, *m3;
    TNormal n;
  public:
    TTriangle(TPoint *p1, TPoint *p2, TPoint *p3):
             n(p1, p2, p3)
    {
      m1 = p1;
      m2 = p2;
      m3 = p3;
    }
    int Visible()
    {
      return (n.z > 0);
    }
    void Draw(unsigned color, unsigned x, unsigned y)
    {
      Triangle2D(*m1, *m2, *m3, x, y, color);
    }
    TNormal Normal()
    {
      return n;
    }
};

class TTetraedr {
    TPoint *p[NUMPOINTS];
    TPoint RotateX(TPoint p, unsigned angle)
    {
      return TPoint(p.x,
                    int((p.y * cos_table[angle] - p.z * sin_table[angle]) >> 16),
                    int((p.y * sin_table[angle] + p.z * cos_table[angle]) >> 16));
    }
    TPoint RotateY(TPoint p, unsigned angle)
    {
      return TPoint(int((p.x * cos_table[angle] - p.z * sin_table[angle]) >> 16),
                    p.y,
                    int((p.x * sin_table[angle] + p.z * cos_table[angle]) >> 16));
    }
    TPoint RotateZ(TPoint p, unsigned angle)
    {
      return TPoint(int((p.x * cos_table[angle] - p.y * sin_table[angle]) >> 16),
                    int((p.x * sin_table[angle] + p.y * cos_table[angle]) >> 16),
                    p.z);
    }
  public:
    TTetraedr(int size)
    {
      p[0] = new TPoint(-size, -size, -size);
      p[1] = new TPoint(-size, size, -size);
      p[2] = new TPoint(size, size, -size);
      p[3] = new TPoint(size, -size, -size);
      p[4] = new TPoint(0, 0, size);
    }
    void Draw(unsigned, unsigned, unsigned, unsigned, unsigned, unsigned);
    ~TTetraedr();
};

void TTetraedr::Draw(unsigned angle_x, unsigned angle_y,
                     unsigned angle_z, unsigned x,
                     unsigned y, unsigned color)
{
  TTriangle *t[NUMPOINTS + 1];
  TPoint *p2[NUMPOINTS];
  int i = 0, offs;

  for (; i < NUMPOINTS; i++)
    p2[i] = new TPoint(RotateX(RotateY(RotateZ(*p[i], angle_z), angle_y), angle_x));

  t[0] = new TTriangle(p2[0], p2[4], p2[1]);
  t[1] = new TTriangle(p2[1], p2[4], p2[2]);
  t[2] = new TTriangle(p2[2], p2[4], p2[3]);
  t[3] = new TTriangle(p2[3], p2[4], p2[0]);
  t[4] = new TTriangle(p2[0], p2[1], p2[2]);
  t[5] = new TTriangle(p2[0], p2[2], p2[3]);

  for (i = 0; i < NUMPOINTS + 1; i++)
  {
    if (t[i] -> Visible())
    {
      offs = unsigned(((t[i] -> Normal()).z * 63) >> 16);
      t[i] -> Draw((color << 6) + offs, x, y);
    }
    delete t[i];
  }

  for (i = 0; i < NUMPOINTS; i++)
    delete p2[i];
}

TTetraedr::~TTetraedr()
{
  for (int i = 0; i < NUMPOINTS; i++)
    delete p[i];
}

class TFlyer {
    int x, y;
    int a_x, a_y, a_z;
    unsigned size, color;
    int dx, dy, d_ax, d_ay, d_az;
    TTetraedr t;
  public:
    TFlyer(unsigned size, int dx, int dy, int d_ax, int d_ay, int d_az,
           unsigned color):
          t(size)
    {
      TFlyer::size = size;
      TFlyer::color = color;
      TFlyer::dx = dx;
      TFlyer::dy = dy;
      TFlyer::d_ax = d_ax;
      TFlyer::d_ay = d_ay;
      TFlyer::d_az = d_az;
      a_x = a_y = a_z = 0;
      x = 160;
      y = 100;
    }
    void Move()
    {
      t.Draw(a_x, a_y, a_z, x, y, color);

      x += dx;
      y += dy;

      a_x += d_ax;
      a_y += d_ay;
      a_z += d_az;

      if ((x > 320 - size * 2) || (x < size * 2)) dx = -dx;
      if ((y > 200 - size * 2) || (y < size * 2)) dy = -dy;

      if (a_x < 0) a_x = 360 + a_x;
      if (a_x > 359) a_x = 0;
      if (a_y < 0) a_y = 360 + a_y;
      if (a_y > 359) a_y = 0;
      if (a_z < 0) a_z = 360 + a_z;
      if (a_z > 359) a_z = 0;
    }
};

void SetGraphMode()
{
  asm  mov    ax,  13h
  asm  int    10h
  asm  mov    dx,  3C4h
  asm  mov    al,  4
  asm  out    dx,  al
  asm  inc    dx
  asm  in     al,  dx
  asm  or     al,  4
  asm  and    al,  0F7h
  asm  out    dx,  al
  asm  mov    dx,  3D4h
  asm  mov    al,  14h
  asm  out    dx,  al
  asm  inc    dx
  asm  in     al,  dx
  asm  and    al,  0BFh
  asm  out    dx,  al
  asm  dec    dx
  asm  mov    al,  17h
  asm  out    dx,  al
  asm  inc    dx
  asm  in     al,  dx
  asm  or     al,  40h
  asm  out    dx,  al
  asm  push   0A000h
  asm  pop    es
  asm  xor    di,  di
  asm  xor    ax,  ax
  asm  mov    cx,  8000h
  asm  cld
  asm  rep    stosw
}

void SetTextMode()
{
  asm  mov    ax,  03h
  asm  int    10h
}

void ClearPage(unsigned page)
{
  asm  push   0A000h
  asm  pop    es
  asm  mov    di,  page
  asm  xor    ax,  ax
  asm  mov    cx,  8192
  asm  mov    dx,  3C4h
  asm  mov    ax,  0F02h
  asm  out    dx,  ax
  asm  cld
  asm  mov    cx,  8192
  asm  xor    ax,  ax
  asm  rep    stosw
}

void TilePage(unsigned page, char color)
{
  asm  push   0A000h
  asm  pop    es
  asm  mov    di,  page
  asm  mov    dx,  3C4h
  asm  mov    cx,  50
  asm  mov    bl,  color
  asm  mov    bh,  color
  asm  cld
l:asm  push   cx
  asm  mov    ax,  0F02h
  asm  out    dx,  ax
  asm  mov    cx,  40
  asm  mov    ax,  bx
  asm  rep    stosw
  asm  push   di
  asm  mov    ax,  0E02h
  asm  out    dx,  ax
  asm  mov    cx,  40
  asm  xor    ax,  ax
  asm  rep    stosw
  asm  mov    cx,  40
  asm  rep    stosw
  asm  mov    cx,  40
  asm  rep    stosw
  asm  pop    di
  asm  mov    ax,  0102h
  asm  out    dx,  ax
  asm  mov    cx,  40
  asm  mov    ax,  bx
  asm  rep    stosw
  asm  mov    cx,  40
  asm  rep    stosw
  asm  mov    cx,  40
  asm  rep    stosw
  asm  pop    cx
  asm  loop   l
}

void SetPage(unsigned page)
{
  asm  mov    bx,  page
  asm  mov    dx,  3D4h
  asm  mov    al,  0Ch
  asm  mov    ah,  bh
  asm  out    dx,  ax
  asm  mov    al,  0Dh
  asm  xor    ah,  ah
  asm  out    dx,  ax
}

void Vertret()
{
    asm  mov    dx,  03DAh
l1: asm  in     al,  dx
    asm  test   al,  8
    asm  jz     l1
l2: asm  in     al,  dx
    asm  test   al,  8
    asm  jnz    l2
}

void SetColor(char color, char r, char g, char b)
{
  asm  mov    dx,  3C8h
  asm  mov    al,  color
  asm  out    dx,  al
  asm  inc    dx
  asm  mov    al,  r
  asm  out    dx,  al
  asm  mov    al,  g
  asm  out    dx,  al
  asm  mov    al,  b
  asm  out    dx,  al
}

void SetPalette()
{
  for (int i = 0; i < 64; i++)
  {
    SetColor(i, 0, 0, i);
    SetColor(i + 64, 0, i, 0);
    SetColor(i + 128, i, 0, 0);
    SetColor(i + 192, i, i, 0);
  }
}

void Message()
{
  cprintf("3D transformations - by HackMaster\n");
}

void main()
{
  SetGraphMode();
  SetPalette();
  TFlyer f1(40, 2, 2, 2, 0, 2, 2);
  TFlyer f2(30, -1, -1, 0, 2, -2, 1);

  do
  {
    SetPage(page ^ 0x4000);
    Vertret();
    TilePage(page, 0x10);
    f2.Move();
    f1.Move();
    page ^= 0x4000;
  } while (!kbhit());

  SetTextMode();
  Message();
}
